import inspect
import sys

from getgauge.registry import registry, _MessagesStore


def step(step_text):
    def _step(func):
        f_back = sys._getframe().f_back
        registry.add_step_definition(step_text, func, f_back.f_code.co_filename, inspect.getsourcelines(func)[1])
        return func

    return _step


def continue_on_failure(func):
    registry.continue_on_failure(func)
    return func


def before_suite(func):
    registry.add_before_suite(func)
    return func


def after_suite(func):
    registry.add_after_suite(func)
    return func


def before_scenario(obj=None):
    return _define_wrapper(obj, registry.add_before_scenario)


def after_scenario(obj=None):
    return _define_wrapper(obj, registry.add_after_scenario)


def before_spec(obj=None):
    return _define_wrapper(obj, registry.add_before_spec)


def after_spec(obj=None):
    return _define_wrapper(obj, registry.add_after_spec)


def before_step(obj):
    return _define_wrapper(obj, registry.add_before_step)


def after_step(obj):
    return _define_wrapper(obj, registry.add_after_step)


def screenshot(func):
    registry.set_screenshot_provider(func)
    return func


class Table:
    def __init__(self, proto_table):
        self.__headers = proto_table.headers.cells
        self.__rows = proto_table.rows

    @property
    def headers(self):
        return self.__headers

    @property
    def rows(self):
        return [row.cells for row in self.__rows]

    def get_row(self, index):
        return self.__rows[index - 1].cells

    def get_column_values_with_name(self, name):
        index = list(self.__headers).index(name)
        return self.get_column_values_with_index(index + 1)

    def get_column_values_with_index(self, index):
        return [row.cells[index - 1] for row in self.__rows]

    def __str__(self):
        return '{}\n{}'.format(', '.join(self.__headers), '\n'.join([', '.join(x.cells) for x in self.__rows]))

    def __eq__(self, other):
        return self.__str__() == other.__str__()

    def __getitem__(self, i):
        return self.get_row(i + 1)

    def __iter__(self):
        return self.rows.__iter__()


class ExecutionContext:
    def __init__(self, specification, scenario, step):
        self.__step = step
        self.__specification = specification
        self.__scenario = scenario

    @property
    def specification(self):
        return self.__specification

    @property
    def scenario(self):
        return self.__scenario

    @property
    def step(self):
        return self.__step

    def __str__(self):
        return 'ExecutionInfo: {{ {}, {}, {} }}'.format(self.specification.__str__(), self.scenario.__str__(),
                                                        self.step.__str__())

    def __eq__(self, other):
        return self.__str__() == other.__str__()


class Specification:
    def __init__(self, name, file_name, is_failing, tags):
        self.__name = name
        self.__file_name = file_name
        self.__is_failing = is_failing
        self.__tags = tags

    @property
    def name(self):
        return self.__name

    @property
    def file_name(self):
        return self.__file_name

    @property
    def is_failing(self):
        return self.__is_failing

    @property
    def tags(self):
        return self.__tags

    def __str__(self):
        return "Specification: {{ name: {}, is_failing: {}, tags: {}, file_name: {} }}".format(self.name,
                                                                                               str(self.is_failing),
                                                                                               ", ".join(self.tags),
                                                                                               self.file_name)

    def __eq__(self, other):
        return self.__str__() == other.__str__()


class Scenario:
    def __init__(self, name, is_failing, tags):
        self.__name = name
        self.__is_failing = is_failing
        self.__tags = tags

    @property
    def name(self):
        return self.__name

    @property
    def is_failing(self):
        return self.__is_failing

    @property
    def tags(self):
        return self.__tags

    def __str__(self):
        return "Scenario: {{ name: {}, is_failing: {}, tags: {} }}".format(self.name, str(self.is_failing),
                                                                           ", ".join(self.tags))

    def __eq__(self, other):
        return self.__str__() == other.__str__()


class Step:
    def __init__(self, text, is_failing):
        self.__text = text
        self.__is_failing = is_failing

    @property
    def text(self):
        return self.__text

    @property
    def is_failing(self):
        return self.__is_failing

    def __str__(self):
        return "Step: {{ text: {}, is_failing: {} }}".format(self.text, str(self.is_failing))

    def __eq__(self, other):
        return self.__str__() == other.__str__()


class Messages:
    @staticmethod
    def write_message(message):
        _MessagesStore.write_message(message)


class DataStore:
    def __init__(self):
        self.__data_store = {}

    def get(self, key):
        return self.__data_store[key]

    def put(self, key, value):
        self.__data_store[key] = value

    def is_present(self, key):
        return key in self.__data_store

    def clear(self):
        self.__data_store = {}

    def __eq__(self, other):
        return self.__data_store == other.__data_store


class DataStoreFactory:
    __scenario_data_store = DataStore()
    __spec_data_store = DataStore()
    __suite_data_store = DataStore()

    @staticmethod
    def scenario_data_store():
        return DataStoreFactory.__scenario_data_store

    @staticmethod
    def spec_data_store():
        return DataStoreFactory.__spec_data_store

    @staticmethod
    def suite_data_store():
        return DataStoreFactory.__suite_data_store


def create_execution_context_from(current_execution_info):
    return ExecutionContext(
        Specification(current_execution_info.currentSpec.name, current_execution_info.currentSpec.fileName,
                      current_execution_info.currentSpec.isFailed, current_execution_info.currentSpec.tags),
        Scenario(current_execution_info.currentScenario.name, current_execution_info.currentScenario.isFailed,
                 current_execution_info.currentScenario.tags),
        Step(current_execution_info.currentStep.step.actualStepText, current_execution_info.currentStep.isFailed)
    )


def _wrapper(*args, **kwargs):
    pass


def _define_wrapper(obj, callback):
    if hasattr(obj, '__call__'):
        callback(obj, None)
        return obj

    def func(function):
        callback(function, obj)
        return function

    return func



import os
import re
import tempfile
from subprocess import call

from getgauge.api import get_step_value


class StepInfo(object):
    def __init__(self, step_text, parsed_step_text, impl, file_name, line_number, has_alias=False):
        self.__step_text = step_text
        self.__parsed_step_text = parsed_step_text
        self.__impl = impl
        self.__file_name = file_name
        self.__line_number = line_number
        self.__has_alias = has_alias

    @property
    def step_text(self):
        return self.__step_text

    @property
    def parsed_step_text(self):
        return self.__parsed_step_text

    @property
    def impl(self):
        return self.__impl

    @property
    def has_alias(self):
        return self.__has_alias

    @property
    def file_name(self):
        return self.__file_name

    @property
    def line_number(self):
        return self.__line_number


class _MessagesStore:
    __messages = []

    @staticmethod
    def pending_messages():
        messages = _MessagesStore.__messages
        _MessagesStore.__messages = []
        return messages

    @staticmethod
    def write_message(message):
        _MessagesStore.__messages.append(message)

    @staticmethod
    def clear():
        _MessagesStore.__messages = []


def _take_screenshot():
    temp_file = os.path.join(tempfile.gettempdir(), 'screenshot.png')
    call(['gauge_screenshot', temp_file])
    _file = open(temp_file, 'r+b')
    data = _file.read()
    _file.close()
    return data


class Registry(object):
    hooks = ['before_step', 'after_step', 'before_scenario', 'after_scenario', 'before_spec', 'after_spec',
             'before_suite', 'after_suite']

    def __init__(self):
        self.__screenshot_provider = _take_screenshot
        self.__steps_map = {}
        self.__continue_on_failures = []
        for hook in Registry.hooks:
            self.def_hook_methods(hook)

    def def_hook_methods(self, hook):
        def get(self, tags=None):
            return _filter_hooks(tags, getattr(self, '__{}'.format(hook)))

        def add(self, func, tags=None):
            getattr(self, '__{}'.format(hook)).append({'tags': tags, 'func': func})

        setattr(self.__class__, hook, get)
        setattr(self.__class__, 'add_{}'.format(hook), add)
        setattr(self, '__{}'.format(hook), [])

    def add_step_definition(self, step_text, func, file_name, line_number=-1, has_alias=False):
        if not isinstance(step_text, list):
            parsed_step_text = get_step_value(step_text)
            self.__steps_map.setdefault(parsed_step_text, []).append(StepInfo(step_text, parsed_step_text, func, file_name, line_number, has_alias))
            return
        for text in step_text:
            self.add_step_definition(text, func, file_name, line_number, True)

    def all_steps(self):
        return [value[0].step_text for value in self.__steps_map.values()]

    def is_step_implemented(self, step_text):
        return self.__steps_map.get(step_text) is not None

    def has_multiple_impls(self, step_text):
        return len(self.__steps_map.get(step_text)) > 1

    def get_info(self, step_text):
        info = self.__steps_map.get(step_text)
        return info[0] if info is not None else StepInfo(None, None, None, None, None)

    def get_infos(self, step_text):
        return self.__steps_map.get(step_text)

    def set_screenshot_provider(self, func):
        self.__screenshot_provider = func

    def screenshot_provider(self):
        return self.__screenshot_provider

    def continue_on_failure(self, func):
        self.__continue_on_failures.append(func)

    def is_continue_on_failure(self, func):
        return func in self.__continue_on_failures

    def clear(self):
        self.__steps_map = {}
        self.__continue_on_failures = []
        for hook in Registry.hooks:
            setattr(self, '__{}'.format(hook), [])


registry = Registry()


def _filter_hooks(tags, hooks):
    filtered_hooks = []
    for hook in hooks:
        hook_tags = hook['tags']
        if hook_tags is None:
            filtered_hooks.append(hook['func'])
            continue
        for tag in tags:
            hook_tags = hook_tags.replace('<{}>'.format(tag), 'True')
        if eval(re.sub('<[^<]+?>', 'False', hook_tags)):
            filtered_hooks.append(hook['func'])
    return filtered_hooks



import re

import sys
from redbaron import RedBaron

from getgauge.registry import registry


def _refactor_content(content, info, request):
    red = RedBaron(content)
    for func in red.find_all('def'):
        for decorator in func.decorators:
            steps = re.findall(r'[\'"](.*?)[\'"]', decorator.call.__str__())
            if len(steps) > 0 and steps[0] == info.step_text:
                refactor_impl(decorator, func, request)
    return red.dumps()


def refactor_impl(decorator, func, request):
    decorator.call = '("' + request.refactorRequest.newStepValue.parameterizedStepValue + '")'
    params = [''] * len(request.refactorRequest.newStepValue.parameters)
    for index, position in enumerate(request.refactorRequest.paramPositions):
        if position.oldPosition < 0:
            params[position.newPosition] = get_param_name("arg{}_".format(position.newPosition), request.refactorRequest.newStepValue.parameters[index])
        else:
            params[position.newPosition] = func.arguments[position.oldPosition].__str__()
    func.arguments = ', '.join(params)


def get_param_name(prefix, param_name):
    if sys.version_info < (3, 0):
        return prefix + "".join([s for s in param_name if re.match(r'^[a-z_][a-z_]*$', s, re.I) is not None or s.isalnum()])
    return prefix + "".join([s for s in param_name if s.isidentifier() or s.isalnum()])


def refactor_step(request, response):
    if registry.has_multiple_impls(request.refactorRequest.oldStepValue.stepValue):
        raise Exception('Multiple Implementation found for `{}`'.format(request.refactorRequest.oldStepValue.parameterizedStepValue))
    info = registry.get_info(request.refactorRequest.oldStepValue.stepValue)
    impl_file = open(info.file_name, 'r+')
    content = _refactor_content(impl_file.read(), info, request)
    impl_file.seek(0)
    impl_file.truncate()
    impl_file.write(content)
    impl_file.close()
    response.refactorResponse.success = True
    response.refactorResponse.filesChanged.append(info.file_name)



from distutils.core import setup
setup(
    name='getgauge',
    packages=['getgauge', 'getgauge/messages'],
    version='0.1.4',
    description='Enables Python support for Gauge',
    author='Kashish Munjal',
    author_email='kashishmunjal64@gmail.com',
    url='https://github.com/kashishm/gauge-python',
    download_url='https://github.com/kashishm/gauge-python/archive/v0.1.4.zip',
    keywords=['testing', 'gauge', 'gauge-python', 'getgauge', 'automation'],
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    install_requires=['protobuf==3.0.0b2.post2', 'redBaron'],
)
